/*  ---------------------------------------------------------------
    xhkeys. Jul 2002
    Copyright (C) 2002,  Michael Glickman  <wmalms@yahoo.com>
    License: GPL
    --------------------------------------------------------------- */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>

#include "xhkeys_conf.h"

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#define _GNU_SOURCE
#include <getopt.h>

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "xhkeys.h"


#define GlobalResDir "/usr/lib/X11/app-defaults"

#define DefResFileName  "XHkeys"
#define ResClassName  "XHkeys"
#define ResModuleName "xhkeys"

CmdDescr *Commands = NULL;
int CommandCount;

int keyTimeout;
int btnTimeout;
int maxCodes;
int maxCodesOriginal;

#ifdef CDROM_SUPPORT
#include <linux/soundcard.h>
char *CDDevName = NULL;
Bool holdCD;
#endif

#ifdef MIXER_SUPPORT
char *MixerDevName = NULL;
int mixerChannel;
Bool holdMixer;
#endif

unsigned long keyLattency, keyPause;
unsigned long clickLattency, clickPause;


const char *CommandTypeCodes = "IAKM";
const char *TargetSelCodes = "CTIO";
const char *KeyTypeModif = "YS";

const char *LogModeCodes = "NEA";
int logMode;		// 0 - none, 1 - error, 2 - all


static Status GetResourceString(XrmDatabase database, const char *parmname,
                              char **string_return)
{

  XrmValue xrmv;
/*  int sz; */
  char *str; 

  if (!XrmGetResource(database, parmname, ResClassName,  &str, &xrmv))
		return False;
/*
  sz = xrmv.size;
  str = malloc(sz+1);
    
  if (str == NULL) return False;

  memcpy(str, xrmv.addr, sz);
  *(str + sz) = '\0';

  *string_return = str;
*/
  return ((*string_return=strdup(xrmv.addr)) != NULL);
}

static Status GetResourceInt(XrmDatabase database, const char *parmname,
                         int min, int max, int *value_return)
{
  char *strPtr;
  int value;
  XrmValue xrmv;

  if (!XrmGetResource(database, parmname, ResClassName,  &strPtr, &xrmv))
		return False;

  value = (int) strtol(xrmv.addr, &strPtr, 0);
  if ((*strPtr == '\0' || *strPtr==' ') && value>=min && value<=max) {
	*value_return = value;
	return True;
  }
  
  return False;
}						 

static Status GetResourceUlong(XrmDatabase database, const char *parmname,
                         long min, long max, unsigned long *value_return)
{
  char *strPtr;
  long value;
  XrmValue xrmv;

  if (!XrmGetResource(database, parmname, ResClassName,  &strPtr, &xrmv))
		return False;

  value = strtoul(xrmv.addr, &strPtr, 0);
  if ((*strPtr == '\0' || *strPtr==' ') && value>=min && value<=max) {
	*value_return = value;
	return True;
  }
  
  return False;
}						 

static Status GetResourceBool(XrmDatabase database, const char *parmname,
                            Bool *value_return)
{
  char *strPtr, c;
  XrmValue xrmv;

  if (!XrmGetResource(database, parmname, ResClassName,  &strPtr, &xrmv))
		return False;

  c = *(xrmv.addr);

  if (strchr("yY1", c)) {
	*value_return = True;
	return True;
  } 
  if (strchr("nN0", c)) {
	*value_return = False;
	return True;
  } 

  return False;
}						 


static XrmDatabase GetResourceDatabase(const char *fileName, Bool getGlobal, char **fullResFileNamePtr)
{
  XrmDatabase  xrmdb = NULL;
  const char *homedir;
  char *FullResFileName = NULL;
  
  maxCodesOriginal = -1;

  if (getGlobal) { 
	xrmdb = XrmGetFileDatabase(GlobalResDir "/" DefResFileName);
	if (xrmdb != NULL)
	  GetResourceInt(xrmdb, ResModuleName ".maxcodes", 1, 255, (int *) &maxCodesOriginal); 
  }	

  if (fileName == NULL) fileName = "~/." DefResFileName;

  if (memcmp(fileName, "~/", 2) == 0) {
	if ((homedir = getenv("HOME")) != NULL &&
	    (FullResFileName = malloc(strlen(homedir) + strlen(fileName))) != NULL)
	{	
	  sprintf (FullResFileName, "%s/%s", homedir, fileName+2);
	  fileName = FullResFileName;
	}	            
	else
	  fileName = NULL;
  }	  

  XrmCombineFileDatabase(fileName, &xrmdb, True);
  if (xrmdb != NULL) {
	int maxCodes1;
    GetResourceInt(xrmdb, ResModuleName ".maxcodes", 1, 255, (int *) &maxCodes1); 
	if (maxCodes1 > maxCodesOriginal) maxCodesOriginal = maxCodes1;
  }
  
  if (maxCodesOriginal < 0) maxCodesOriginal = 50;	

  *fullResFileNamePtr = FullResFileName;  
  return xrmdb;
}


Bool ProcessResources(const char *fileName, Bool getGlobal, Bool getCmdLines, char **errMsg)
{
  XrmDatabase  xrmdb = NULL; 
  int  i, j, cmdType;
  char *FullResFileName = NULL;
  char *parm;
  char *codeStr, *modifStr, *typeStr, *comPtr, *cmdNext;
  char parmname[81];
  short type_modifier;	      // Key: 0 - keysym (Y), 1 - scan code (S)
  short target_selection;	  // Select target window by:
                              //    0 - WM_CLASS (C), 1 - Title (T), 2 - ID (I), 3 - Stacking Order(0)
  char targetSelCode;
  Bool success = False;

#ifdef CDROM_SUPPORT
  CDDevName = strdup("/dev/cdrom");
  holdCD = False;
#endif  

#ifdef MIXER_SUPPORT
  MixerDevName = strdup("/dev/mixer");
  holdMixer = False;
  mixerChannel = -1;
#endif  


  keyTimeout = 5;	/* 5 secs */
  btnTimeout = 8;	/* 8 secs */
  keyLattency = 0;
  keyPause = 50000;		/* 50 millisec */
  clickLattency = 0;
  clickPause = 50000;	/* 50 millisec */


  XrmInitialize();
  xrmdb = GetResourceDatabase(fileName, getGlobal, &FullResFileName);

  if (xrmdb == NULL) {
	*errMsg = "Configuration file not found";
	goto OutOfHere;
  }

  maxCodes = maxCodesOriginal;

  GetResourceInt(xrmdb, ResModuleName ".keytimeout", 1, 600, &keyTimeout);
  GetResourceInt(xrmdb, ResModuleName ".clicktimeout", 1, 600, &btnTimeout);

#ifdef CDROM_SUPPORT
  GetResourceString(xrmdb, ResModuleName ".cddevice", &CDDevName); 
  GetResourceBool(xrmdb, ResModuleName ".cdhold", &holdCD);
#endif  


#ifdef MIXER_SUPPORT
  GetResourceString(xrmdb, ResModuleName ".mixerdevice", &MixerDevName); 
  GetResourceInt(xrmdb, ResModuleName ".mixerchannel", -1, SOUND_MIXER_NRDEVICES-1, &mixerChannel);
  GetResourceBool(xrmdb, ResModuleName ".mixerhold", &holdMixer);
#endif  

  GetResourceUlong(xrmdb, ResModuleName ".keylattency", 0, 100000000l,  &keyLattency);
  GetResourceUlong(xrmdb, ResModuleName ".keypause", 0, 100000000l,  &keyPause);
  GetResourceUlong(xrmdb, ResModuleName ".clicklattency", 0, 100000000l,  &clickLattency);
  GetResourceUlong(xrmdb, ResModuleName ".clickpause", 0, 100000000l,  &clickPause);
  

  if (logMode < 0 &&
	GetResourceString(xrmdb, ResModuleName ".logmode", &parm)) {
    parm = strchr(LogModeCodes, toupper(*parm));
	if (parm != NULL) logMode = (parm-LogModeCodes);
  }	

  if (getCmdLines == False) {
	success = True;
	goto OutOfHere;
  }	


  CommandCount = 0;

  
  Commands = malloc(maxCodes * sizeof(CmdDescr));
  if (Commands == NULL) {
	*errMsg = "MaxCodes is too big";
	goto OutOfHere;
  }
 
  for (i=1; i<=maxCodes; i++)  {
	sprintf(parmname,  ResModuleName ".codeline%d", i);

    if (GetResourceString(xrmdb, parmname, &parm) && *parm != '\0') {
	  cmdNext = parm;	  
	  codeStr = strsep(&cmdNext, ";");
	  modifStr = strsep(&cmdNext, ";");
	  typeStr = strsep(&cmdNext, ";");
	  comPtr = strsep(&cmdNext, "");
	  if (codeStr != NULL && modifStr != NULL && typeStr != NULL && comPtr != NULL) {
		cmdType = *typeStr++;
	
		if (isdigit(cmdType)) {
		  cmdType -= '0';  // cmdType = atoi(typeStr);
		  if (cmdType >= CommandTypeCount) cmdType = -1;
		}  
		else {
		  cmdNext = strchr(CommandTypeCodes, cmdType);
		  if (cmdNext == NULL) cmdType = -1;
		  else cmdType = (cmdNext-CommandTypeCodes);
		}

		if (cmdType >= 0) {		  
		  type_modifier = 0;	  	// By keysym					   
		  if (strchr(typeStr, 'S')) type_modifier = 1;        // By scancode

		  target_selection = 0;	// By WM_CLASS
		  for (j=1; (targetSelCode = TargetSelCodes[j]) != '\0'; j++) {
			if (strchr(typeStr, targetSelCode)) {
				target_selection = j; break;
			}	
		  }				

		  Commands[CommandCount].lineno = i;
		  Commands[CommandCount].keycode = (unsigned int) strtoul(codeStr, NULL, 0);
		  Commands[CommandCount].modifier = (unsigned int) strtoul(modifStr, NULL, 0);
  		  Commands[CommandCount].type = cmdType;
		  Commands[CommandCount].type_modifier = type_modifier;
		  Commands[CommandCount].target_selection = target_selection;
		  Commands[CommandCount].command = strdup(comPtr);
		  
		  CommandCount++;
		}  
	  }	
	  free(parm);
	}
  }

  success = True;
  

OutOfHere:
  if (FullResFileName != NULL) free(FullResFileName);
  if (xrmdb != NULL) XrmDestroyDatabase(xrmdb);
  return success;
}


Bool UpdateResource(const char *fileName, int pos, char **errMsg)
{
  char line[301];
  char resName[121];
  char typeChr[5];
  XrmDatabase  xrmdb = NULL; 
  int  j, cmdType; 
  char *FullResFileName = NULL;
  Bool success = False;
  CmdDescr *cmd;

  xrmdb = GetResourceDatabase(fileName, False, &FullResFileName);

/*
  if (xrmdb == NULL) {
	*errMsg = "Configuration file not found";
	goto OutOfHere;
  }
*/

  sprintf(line, "%d", maxCodes);
  XrmPutStringResource(&xrmdb, ResModuleName ".maxcodes", line);

  cmd = &Commands[pos]; 
  sprintf(resName, ResModuleName ".codeline%d", cmd->lineno);

  cmdType = cmd->type;
  if(cmdType == CT_DELETED)
	*line = '\0';
  else {	
	j = 0;
	typeChr[j++] = CommandTypeCodes[cmdType];
	if (cmdType == CT_KEYEVENT || cmdType == CT_BTNEVENT)
	  typeChr[j++] = TargetSelCodes[cmd->target_selection];
	if (cmdType == CT_KEYEVENT)
	  typeChr[j++] = KeyTypeModif[cmd->type_modifier];
	typeChr[j] = '\0';

	sprintf(line, "%d;%d;%s;%s", cmd->keycode, cmd->modifier, typeChr, cmd->command);		
  }	
  XrmPutStringResource(&xrmdb, resName, line);


  XrmPutFileDatabase(xrmdb, FullResFileName);
  if (maxCodes > maxCodesOriginal)
	  maxCodesOriginal = maxCodes;
  success = True;

/* OutOfHere: */
  if (FullResFileName != NULL) free(FullResFileName);
  if (xrmdb != NULL) XrmDestroyDatabase(xrmdb);
  return success;
}


void FreeResources(void)
{
  int i;
  char *cmd;

#ifdef CDROM_SUPPORT
  if (CDDevName != NULL) {
	free(CDDevName);
	CDDevName = NULL;
  }	
#endif
#ifdef MIXER_SUPPORT
  if (MixerDevName != NULL) {
	free(MixerDevName);
	MixerDevName = NULL;
  }	
#endif

  if (Commands == NULL) return;
  
  for (i=0; i<CommandCount; i++)  {
	if ((cmd=(char *)Commands[i].command) != NULL) free(cmd);
  }
  
  free(Commands); Commands = NULL;
}
