/*  ---------------------------------------------------------------
    xhkeys. Jul 2002
    Copyright (C) 2002,  Michael Glickman  <wmalms@yahoo.com>
    License: GPL
    --------------------------------------------------------------- */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xhkeys_conf.h"

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef CDROM_SUPPORT
# include <linux/cdrom.h>
#endif

#ifdef MIXER_SUPPORT
# include <linux/soundcard.h>
#endif

#if defined(CDROM_SUPPORT) || defined(MIXER_SUPPORT)
# include <sys/ioctl.h>
# include <fcntl.h>
# include <sys/stat.h>
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "xhkeys.h"

// system.c
extern char **environ;

// xhkeys.c
extern Display *dpy;
extern Window wnd;

// resources.c
extern const char *SpecCodeNames[];
extern const int SpecCodeCount;

#ifdef CDROM_SUPPORT
extern char *CDDevName;
static int cdrom_fd;
extern Bool holdCD;
#endif

#ifdef MIXER_SUPPORT
extern char *MixerDevName;
static int mixer_fd;
extern int mixerChannel;
extern Bool holdMixer;
#endif

#define MAX_VOLUME 100
#define MIN_VOLUME 0

// common.c
extern const char *SpecCodeNames[];
extern const int SpecCodeCount;


static Bool parse_parm_plusminus(char *parm, int *valPtr, short *dirPtr, char **sepPtr)
{
  short dir = 0;
  int val = 0;
  char *sep;
  
  while (*parm == ' ') parm++;
  
  if (*parm == '+') { dir = 1; parm++; }
  if (*parm == '-') { dir = -1; parm++; }
  if (*parm == '=') { dir = 2; parm++; }

  val = strtol(parm, &sep, 0);
  if (sepPtr == NULL) {
    if (*sep != ' ' && *sep != '\0') return False;
  } else
	*sepPtr=sep;
  
  *valPtr = val;
  *dirPtr = dir;
  return True;	
}

#ifdef MIXER_SUPPORT

static Bool get_mixer_volumes(int *devPtr, int *volumes)
{
  static const int devices[] = {SOUND_MIXER_VOLUME, SOUND_MIXER_PCM, SOUND_MIXER_ALTPCM};
  int device = *devPtr;
  int i;
  Bool success;

  if (device >= 0)
      success =(ioctl(mixer_fd, MIXER_READ(device), volumes) != -1);
  else {	  
	  for (i=0;
    	 i<3 && ioctl(mixer_fd, MIXER_READ(device=devices[i]), volumes) == -1;
		 i++);
    success = (i < 3);
  }
  
  if (success) *devPtr = device;
  return success;	
}

static Bool set_mixer_volumes(int device, int volLeft, int volRight)
{
  if (volLeft < MIN_VOLUME) volLeft=MIN_VOLUME;
  else
  if (volLeft > MAX_VOLUME) volLeft=MAX_VOLUME;
  
  if (volRight < MIN_VOLUME) volRight=MIN_VOLUME;
  else
  if (volRight > MAX_VOLUME) volRight=MAX_VOLUME;
    
  volLeft |= (volRight<<8);  
  return (ioctl(mixer_fd, MIXER_WRITE(device), &volLeft) != -1);
}

static Bool mixer_volume(char *parm)
{
  Bool success = False;
  int volLeft, volRight, value;
  short dir;
  char *chanTxt, *nextTxt;
  int device=0; 

  if (parse_parm_plusminus(parm, &value, &dir, &chanTxt) == False)
	return False;

  device = mixerChannel;

  if (strchr(";.,-", *chanTxt)) {
	volLeft = strtol(++chanTxt, &nextTxt, 0);
    if ((*nextTxt == '\0' || *nextTxt == ' ') &&
	    volLeft < SOUND_MIXER_NRDEVICES)  device = volLeft;
  }	  

  if (mixer_fd < 0)  
	mixer_fd = open(MixerDevName, O_RDWR);

  if (mixer_fd < 0) goto OutOfHere;
  
  if (get_mixer_volumes(&device, &volLeft) == False)
    goto OutOfHere;
  
  if (dir == 2) 
	volRight = volLeft = value;
  else {
    volRight = (volLeft>>8) & 0xff;
	volLeft &= 0xff;
  	
	if (dir == 0) 
	  value -= (volLeft+volRight) / 2;
	else
	if (dir < 0)
	  value = -value;
	  
	volLeft += value;  volRight += value;	  	  
  }	
  
  success = set_mixer_volumes(device, volLeft, volRight);
  

OutOfHere:
  if (mixer_fd >= 0 && holdMixer == False) {
	close(mixer_fd);
	mixer_fd = -1;
  }	
  return success;

}

static Bool mixer_balance(char *parm)
{
  Bool success = False;
  int volLeft, volRight, value;
  int volume, balance;
  short dir;
  char *chanTxt, *nextTxt;
  int device=0; 

  
  if (parse_parm_plusminus(parm, &value, &dir, &chanTxt) == False)
	return False;
  if (dir == 2) dir = 0;

  device = mixerChannel;
  if (strchr(";.,-", *chanTxt)) {
	volLeft = strtol(++chanTxt, &nextTxt, 0);
    if ((*nextTxt == '\0' || *nextTxt == ' ') &&
	    volLeft < SOUND_MIXER_NRDEVICES)  device = volLeft;
  }	  

  if (mixer_fd < 0)  
	mixer_fd = open(MixerDevName, O_RDWR);

  if (mixer_fd < 0) goto OutOfHere;
  
  if (get_mixer_volumes(&device, &volLeft) == False)
    goto OutOfHere;
  
  volRight = (volLeft>>8) & 0xff;
  volLeft &= 0xff;

  balance = volRight - volLeft;
  volume = (balance >= 0) ? volRight : volLeft;

  if (dir == 0) 
	balance = value;
  else {
	if (dir < 0) value =-value;
	balance += value;
  }

  if (balance >= 0) {
	volRight = volume;
	volLeft = volume - balance;
  } else {
	volLeft = volume;
	volRight = volume + balance;	// volRight will be actually less
  }
  
  success = set_mixer_volumes(device, volLeft, volRight);

OutOfHere:
  if (mixer_fd >= 0 && holdMixer == False) {
	close(mixer_fd);
	mixer_fd = -1;
  }	
  return success;

}
#endif

#ifdef CDROM_SUPPORT
static void get_cdrom_fd(void)
{
  
  if (cdrom_fd < 0)
	cdrom_fd = open(CDDevName, O_RDONLY | O_NONBLOCK);
   if (cdrom_fd >= 0) {
  	 ioctl(cdrom_fd, CDROMEJECT_SW, NULL);
   }
  	  
}	  

static void release_cdrom_fd(Bool force)
{
  if (cdrom_fd >= 0 && (force || holdCD==False)) {
	close(cdrom_fd);
	cdrom_fd = -1;
  }	
}

//----------------------------------------------------------
// Playing routines
static void lbatomsf(int lba, short *minute, short *second, short *frame)
{
	lba += CD_MSF_OFFSET;
	*frame = lba % CD_FRAMES;
	lba /= CD_FRAMES;
	*second = lba % CD_SECS;
	*minute = lba / CD_SECS;
}

/*
static int msftolba(short minute, short second, short frame)
{
	return (minute*CD_SECS+second)*CD_FRAMES+frame-CD_MSF_OFFSET; 	
}
*/

static Bool cd_play_from_lba(int lba, int dir) {
  struct cdrom_msf playmsf;
  int max_lba;

  {
    struct cdrom_tocentry te;
	te.cdte_track = CDROM_LEADOUT;
	te.cdte_format = CDROM_LBA;
    if(ioctl(cdrom_fd, CDROMREADTOCENTRY, &te) == -1) return False; 
	max_lba = te.cdte_addr.lba;
  }
    
  if (lba < 0 || lba >= max_lba)	return False;
  {  
	short minute, second, frame;
	lbatomsf(lba, &minute, &second, &frame);
	playmsf.cdmsf_min0 = minute;
    playmsf.cdmsf_sec0 = second;
	playmsf.cdmsf_frame0 = frame;
	lbatomsf(max_lba, &minute, &second, &frame);
	playmsf.cdmsf_min1 = minute;
    playmsf.cdmsf_sec1 = second;
	playmsf.cdmsf_frame1 = frame;
  }	

  return (ioctl(cdrom_fd, CDROMPLAYMSF, &playmsf) != -1);
}

static Bool cd_play_from_trk(short trk, short index, short dir) {
  struct cdrom_ti playti;
  struct cdrom_tochdr thdr;

  if (ioctl(cdrom_fd, CDROMREADTOCHDR, &thdr) == 1) return False;

  if (trk < thdr.cdth_trk0)  trk = (dir < 0) ? thdr.cdth_trk1 : thdr.cdth_trk0;
  if (trk > thdr.cdth_trk1)  trk = (dir > 0) ? thdr.cdth_trk0 : thdr.cdth_trk1;;
    
  playti.cdti_trk0 = trk;
  playti.cdti_trk1 = index;
  playti.cdti_trk1 = CDROM_LEADOUT;
  playti.cdti_ind1 = 0;

  return (ioctl(cdrom_fd, CDROMPLAYTRKIND, &playti) != -1);
}

static Bool cd_play_parsed(short addrType, int value, short dir) 
{

  if (addrType == 2)
  {	// Start of disk
	struct cdrom_tochdr thdr;
	if (ioctl(cdrom_fd, CDROMREADTOCHDR, &thdr) == 1) return False;
	return cd_play_from_trk(thdr.cdth_trk0, 0, 1);
  }

  if (dir != 0) {
    struct cdrom_subchnl subch;
	subch.cdsc_format = CDROM_LBA;
	if (ioctl(cdrom_fd, CDROMSUBCHNL, &subch) == -1) return -1;
	if (dir < 0) value = -value;
	value += (addrType == 0) ? (int) subch.cdsc_trk & 0xff : subch.cdsc_absaddr.lba;
  }

  if (addrType == 0)
	  return cd_play_from_trk(value, 0, dir);

  return cd_play_from_lba(value, dir);
  	
}


// Returns the type of value:
//	 0 - TRACK absolute,
//   1 -  LBA absolute,
//   2 - start of disk
//   (-1) - invalid
static short cd_parse_play_parm(char *parm, int *valPtr, short *offsTypePtr)
{
  char *sepPtr;
  int addrType = -1;		// We don't know the type yet
  int offsType = 0;			// 0 -abs addr, 1 - plus, (-1) - minus,
  int value = 0;
  int crnt = 0;

  
  while (*parm == ' ') parm++;
  
  switch(toupper(*parm))
  {
	case '+':
	  offsType = 1; parm++;
	  break;
	  
	case '-':
	  offsType = -1; parm++;
	  break;

	case 'S':
	case 'D':
	  addrType = 2;
	  break;
  
	case '*':			// Current position
	case '\0':
	  addrType = 1;
	  offsType = 1;
	  break;
	  
	default:
	  offsType = 0;
  }	  

  if (addrType != 2)
	while (*parm != '\0') {
  	  crnt = strtol(parm, &sepPtr, 0);

	  switch(toupper(*sepPtr))
	  {
		case 'M':					// minutes
		  crnt *= CD_SECS;		// convert minus to seconds
	                            // no break!  					  

		case 'S':					// seconds 
		  crnt *= CD_FRAMES;		// convert seconds to frames
	                            // no break!  					  

	    case 'F':					// frames
		  if (addrType == 0) return -1;  // No contradiction
		  addrType = 1;			// MSF
	  	  value += crnt;
		  break;	                // At last !		
		
		case 'T':					// tracks
		case '\0':
		  if (addrType == 1) return -1;	 // No contradiction
		  addrType = 0;			// Track
	  	  value = crnt;
		break;

                          // We just use the following break :=)				
	  default:		
		return -1;				// Invalid character	
	}
	
	if (*sepPtr == '\0') break;
	parm = sepPtr+1;	
	while (*parm == ' ') parm++;
  }

  if (offsType == 0 && addrType == 1)
	value -= CD_MSF_OFFSET; 	 // For absolute MSF take away 2 secs

  *offsTypePtr = offsType;
  *valPtr = value;
  return addrType;

}


static Bool cd_play_parm(char *parm)
{
  int value;
  short result, dir;
  Bool success;

  get_cdrom_fd();
  
  if (cdrom_fd < 0) return False;

  result = cd_parse_play_parm(parm, &value, &dir);
  if (result == -1) return False;

  success = cd_play_parsed(result, value, dir);
  
  if (success == False)
	success = cd_play_parsed(2, 0, 1);        // Play from disc start
  
  release_cdrom_fd(False);
  return success;
  
}


static Bool cd_pauseresume(void) {
  Bool success = False;
  
  get_cdrom_fd();

  if (cdrom_fd >= 0) {
	struct cdrom_subchnl subch;
	subch.cdsc_format = CDROM_LBA;
		
	if (ioctl(cdrom_fd, CDROMSUBCHNL, &subch) != -1) {
	  switch(subch.cdsc_audiostatus) {
		case CDROM_AUDIO_PAUSED:
		  success = (ioctl(cdrom_fd, CDROMRESUME) != -1);
		  break;
		  
		case CDROM_AUDIO_PLAY:
		  success = (ioctl(cdrom_fd, CDROMPAUSE) != -1);
		  break;

	  
	  }  // switch (...)

	  if (success == False)
		  success = cd_play_parsed(1, 0, 1);		// Play from current

	  if (success == False)
		  success = cd_play_parsed(2, 0, 1);        // Play from disc start
		  
	}	 // (if (ioctl(....) != -1)
  } // (if (cdrom_fd >= 0)
  
  release_cdrom_fd(False);
  return success;
}

//----------------------------------------------------------
static Bool cd_slot(char *parm)
{
  int val, crnt;
  short dir;
  Bool success = False;
  
  if (parse_parm_plusminus(parm, &val, &dir, NULL) == False) return False;
  if (dir == 2) dir = 0;

  get_cdrom_fd();
  if (cdrom_fd < 0) return False;

  crnt = ioctl(cdrom_fd, CDROM_SELECT_DISC, CDSL_CURRENT);
  if (crnt >= 0) {
	if (dir != 0) {
	  if (dir < 0) val = -val;
	  val += crnt;
	}

    if (val < 0) val = 0;
	else {  
	  int maxval = ioctl(cdrom_fd, CDROM_CHANGER_NSLOTS);
	  if (val >= maxval) val = maxval-1;
    }
  
	success = (ioctl(cdrom_fd, CDROM_SELECT_DISC, val) != -1);
  }	

  release_cdrom_fd(False);
  return success;
	
}

static Bool cd_justioctl(int request, void *parm)
{
  Bool success = False;
  get_cdrom_fd();

  if (cdrom_fd >= 0) {
	success = (ioctl(cdrom_fd, request, parm) != -1);
    release_cdrom_fd(False);
  }	
  
  return success;
}



static Bool cd_eject(void)
{
  Bool success = False;
  get_cdrom_fd();

  if (cdrom_fd >= 0) {
	ioctl(cdrom_fd, CDROM_LOCKDOOR, 0);
	success = (ioctl(cdrom_fd, CDROMEJECT) != -1);
    release_cdrom_fd(True);
  }	

  return success;
}

static Bool cd_loadeject(void)
{
  Bool success = False;
  get_cdrom_fd();

  if (cdrom_fd >= 0) {
    int res = ioctl(cdrom_fd, CDROM_DRIVE_STATUS, CDSL_CURRENT);

    if (res == CDS_DISC_OK || res == CDS_NO_INFO) {
		ioctl(cdrom_fd, CDROM_LOCKDOOR, 0);
		success = (ioctl(cdrom_fd, CDROMEJECT) != -1);
	    release_cdrom_fd(True);
	}	
	else  {
		success = (ioctl(cdrom_fd, CDROMCLOSETRAY) != -1);
	    release_cdrom_fd(False);
	}	

  }

  return success;
}
#endif


static Bool wnd_rotate(Window main_wnd, char *parm)
{
  int value;
  short dir;
  Window cur;

  if (parse_parm_plusminus(parm, &value, &dir, NULL) == False)
  {  value = 1; dir = 1; }

  if (value <= 0) value = 1;
  if (dir >= 0) dir = 1; else dir = -1;

  while (--value >= 0) {
    cur = findVisibleWindow(main_wnd, dir);
	if (cur == None) return False;
	if (dir >= 0) XRaiseWindow(dpy, cur);
	        else  XLowerWindow(dpy, cur);

  }
  
  
  return True;  
}  

static Bool ProcessSpecCommandByCode(int code, char *parm)
{
  Bool success = False;

  switch(code) {
#ifdef CDROM_SUPPORT
	  case SPEC_CDPLAYPAUSE:
		success = cd_pauseresume();
		break;

	  case SPEC_CDPLAY:
		success = cd_play_parm(parm);
		break;

	  case SPEC_CDSTOP:
		success = cd_justioctl(CDROMSTOP, 0);
		break;

	  case SPEC_CDEJECT:
		success = cd_eject();
		break;

	  case SPEC_CDLOADEJECT:
		success = cd_loadeject();
		break;

	  case SPEC_CDLOAD:
		success = cd_justioctl(CDROMCLOSETRAY, 0);
		break;

	  case SPEC_CDSLOT:
		success = cd_slot(parm);
		break;
#endif
#ifdef MIXER_SUPPORT
	  case SPEC_SNDVOLUME:
		success = mixer_volume(parm);
		break;

	  case SPEC_SNDBALANCE:
		success = mixer_balance(parm);
		break;

#endif

	  case SPEC_WNDROTATE:
//	    XCirculateSubwindowsUp(dpy, wnd);
		success = wnd_rotate(wnd, parm);
		break;



  }

  return success;
  
}


Bool InitSpecFunctions(char **errMsg)
{
#ifdef CDROM_SUPPORT
  cdrom_fd = -1;  
#endif
#ifdef MIXER_SUPPORT
  mixer_fd = -1;  
#endif
  return True;

}

void DeinitSpecFunctions(void)
{
#ifdef CDROM_SUPPORT
  if (cdrom_fd >= 0) close (cdrom_fd);  
#endif
#ifdef MIXER_SUPPORT
  if (mixer_fd >= 0) close (mixer_fd);  
#endif
}

// Returns 1-success, 0-fail, (-1)-quit command
int ProcessSpecFunction(const char *command)
{
  char *cmdCopy, *argPtr, *cmdNext;
  int i, rc = 0;

  cmdCopy = strdup(command);
  if (cmdCopy == NULL) goto OutOfHere;

  cmdNext = strpbrk(cmdCopy, "#\t");
  if (cmdNext != NULL) *cmdNext = '\0';

  cmdNext = cmdCopy;
  argPtr = strsep(&cmdNext, " ;,");
  if (argPtr == NULL) goto OutOfHere;  
  
  for (i=0; i<SpecCodeCount && strcasecmp(SpecCodeNames[i], argPtr);  i++);

  if (i >= SpecCodeCount) goto OutOfHere;
  if (i==SPEC_QUIT) { rc=-1; goto OutOfHere; }
  
  argPtr = strsep(&cmdNext, "");
  if (argPtr) { while (*argPtr == ' ') argPtr++; }
  else argPtr = "";
  
  if (ProcessSpecCommandByCode(i, argPtr)) rc = 1;
  
OutOfHere:
  if (cmdCopy) free(cmdCopy);    
  return rc;
}  

