 /* --------------------------------------------------------------- 
    xhkeys. Jul 2002
    Copyright (C) 2002,  Michael Glickman  <wmalms@yahoo.com>
    License: GPL
    --------------------------------------------------------------- */
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "xhkeys_conf.h"

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "xhkeys.h"

const char *cancelledText = "Cancelled";
const char *timeoutText = "Cancelled by timeout";
const char *timeoutBye = "\nQuit on time-out. Bye\n";

// xhconf.c
extern Display *dpy;
extern Window wnd;
extern int keyTimeout, btnTimeout;
extern int count;
extern const char *resFileName;

// resources.c
extern const char *TargetSelCodes;
extern CmdDescr *Commands;
extern int CommandCount;
extern int maxCodes;

const char *CommandTypeNames[] = {"Internal Function",	"Application",
			"Key Event", "Mouse Event"};

extern const char *CommandTypeCodes;

// common.c
extern const char *SpecCodeNames[];
extern const int SpecCodeCount;

const char *GenerateNames[] = {"None", "Press",	"Release",  "Press and Release"};

const char *StackOrderNames[] = {"Topmost", "Bottom",  "Root"};

static Bool displayCommandLine(CmdDescr *cmd, char **errMsg)
{
  KeyCode code;
  unsigned int type, val;
  unsigned long lval;
  
  
  code = cmd->keycode;
  val = cmd->modifier;
  printf ("Scan code: %3u (0x%02x)  State: %3u (0x%02x)\n",
            code, code, val, val); 	  

  type = cmd->type;  
  printf ("\tOperType: %c [%s]\n", CommandTypeCodes[type], CommandTypeNames[type]);


  if (type == CT_INTERNAL ||  type == CT_EXTAPP) 
	printf ("\tCommand:  %s\n", cmd->command);
  else
  {	//  CT_KEYEVENT, CT_MOUSEEVENT
	char *cmdcopy, *strNext;
	char *ptr1, *ptr2;
	
	strNext = cmdcopy = strdup(cmd->command);
	ptr1 = strsep(&strNext, ";");

	switch (cmd->target_selection) {
	  case 0:
		ptr2 = ptr1;
		ptr1 = strsep(&ptr2, ":");
	    if (ptr1 != NULL || ptr2 != NULL) 
		  printf ("\tWnd App Name: %s   WndAppClass: %s\n", ptr1, ptr2);
		break;

	  case 1:
		printf ("\tWnd Title: '%s'\n", ptr1);
		break;
		
	  case 2:
		lval = strtoul(ptr1, (char **) NULL, 0);	  
		printf ("\tWnd ID: %10ld (0x%lx)\n", lval, lval);		
		break;

	  case 3:
	  {
		char sCode = toupper(*ptr1);
		const char *ordName = "Unknown";
		const char *ordnameCurrent;
		int j;
		
		for (j=0; j<3; j++) {
			ordnameCurrent = StackOrderNames[j];
			if (*ordnameCurrent  == sCode) {
				ordName = ordnameCurrent;
				break;
			}
		}		
		printf ("\tWnd Order: %s\n", ordName);
	  }	
		break;
		
		
	}	

	ptr1 = strsep(&strNext, ";");

    if (ptr1 != NULL && type == CT_BTNEVENT)  {
	  ptr2 = strsep(&strNext, ";");
	  if (ptr2 == NULL)
		ptr1 = NULL;
	  else {
		int x = (int) strtol(ptr1, (char **) NULL, 0);
		int y = (int) strtol(ptr2, (char **) NULL, 0);
		printf ("\tWnd Location (x,y):  %d, %d\n", x, y);
	    ptr1 = strsep(&strNext, ";"); 
	  }
	}

	if (ptr1 != NULL) {
	  ptr2 = strsep(&strNext, ";");
	  val = (ptr2 == NULL) ? 0 : atoi(ptr2);

	  if (type == CT_KEYEVENT) {
		if (cmd->type_modifier == 0)
			printf ("\tKeysym: %s  State %u (0x%x)\n", ptr1, val, val);
		else {
		  code = strtol(ptr1, (char **)NULL, 0);
		  printf ("\tScan code: %d (0x%x)  State %u (0x%x)\n", code, code, val, val);
		}
	  }	  
	  if (type == CT_BTNEVENT) {
		  code = strtol(ptr1, (char **)NULL, 0);
		  printf ("\tMouse button: %d   State %u (0x%x)\n", code, val, val);
	  }
	  	  
	  ptr1 = strsep(&strNext, ";");
	}  

	val = (ptr1 == NULL) ? 3 : atoi(ptr1);
	printf ("\tGenerate: %d [%s]", val, GenerateNames[val]);

	if (ptr1 != NULL  && (ptr1 = strsep(&strNext, "")) != NULL &&
	    (val = atoi(ptr1)) > 0)
		  printf ("   %d times", val);
		  
	printf ("\n");	  
  }
  printf ("\n");

  return True;  
}

static Bool displayCommandEntry(int index, char **errMsg)
{
  return displayCommandLine(&(Commands[index]), errMsg);
}

static int findCommandByCode(unsigned int scanCode, unsigned int modif)
{
  int i;
  
  for (i=0; i<CommandCount; i++)
  {
	if (Commands[i].keycode == scanCode &&
	    Commands[i].modifier == modif) return i;
  }
  return -1;
}

Bool displayCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg)
{
  int i = findCommandByCode(scanCode, modif);
  
  if (i>=0)
	return displayCommandEntry(i, errMsg);

  printf ("Scan code: %3u (0x%02x)  State: %3u (0x%02x)\n",
            scanCode, scanCode, modif, modif); 	  
  printf ("\tUnassigned\n\n");
  return True;
}

Bool displayCommand(char **errMsg)
{
  XEvent xev;
  short res;
  short cnt;
  
  printf("Type any key to see the associated command\n");
  printf ("Quits ");
  if (count > 1)  printf("after %d entries, or ", count);
  printf ("if idle within %d secs\n", keyTimeout);
  cnt = count;
  
  while((res = getScanCodeEvent(&xev, keyTimeout, errMsg)) == 1)
  {
	if (displayCommandByCode(xev.xkey.keycode, xev.xkey.state, errMsg) == False)
	  return False;
	if (cnt > 0 && --cnt==0) break;
  }	 

  if (res == 2)
	printf(timeoutBye);
  return (res != 0);	 
}

Bool displayAllCommands(char **errMsg)
{
  int i;
  
  for (i=0; i<CommandCount; i++)
  {
	if (!displayCommandEntry(i, errMsg)) return False;
  }

  return True;

}

/* -------------------------------------------------------------- */

static void EnterText(const char *prompt, char *answer, int len)
{
  char *sep;

  printf (prompt);
  fgets(answer, len, stdin);
  fflush(stdin);
  sep = strpbrk(answer, "\r\n");
  if (sep) *sep = '\0';
}

static Bool YesNo(const char *question, Bool def_answer)
{
  char qcopy[121], reply[5];
  int  repval = -1;
  
  strcpy(qcopy, question);
  strcat(qcopy, def_answer ? "? (y, n): " : "? (n, y): ");
  
  while (repval < 0) {
	EnterText(qcopy, reply, sizeof(reply));

	switch(toupper(*reply)) {
	  case 'Y':  repval = 1; break;
	  case 'N':  repval = 0; break;

	  case '\r':
	  case '\n':
	  case '\0': repval = (int)def_answer;
	}	
  }	  

  return (Bool) repval; 
}

static int DisplayMenu(const char *title, const char **lines, int count)
{
  int i, iter;
  char answer[5], *ansnext;

  printf ("\n%s\n", title);
  iter = 3;
  
  while(--iter >= 0) {
	for (i=0; i<count; i++)
	  printf ("%d. %s\n", i+1, lines[i]);
  	
	EnterText("Enter selection: ", answer, sizeof(answer));
	i = (int) strtol(answer, &ansnext, 10);
	if (i > 0 && i <= count && *ansnext == '\0')      
	  return i-1;
  }

  return -1;  
}

Bool editSpecFunction(char *command, int sze, char **errMsg)
{
  int fcode;
  char parms[41];
  int iter;
  
  iter = 3;
  
  do {
	EnterText("Function Code (H-help): ", command, sze);

	if (command[0] == '\0') {
	  *errMsg = (char *) cancelledText ;
	   return False;
	}
	
	if (toupper(command[0]) == 'H') 
	  fcode = DisplayMenu("Select finction", SpecCodeNames, SpecCodeCount);
	else {
	  for (fcode = 0; 
		fcode < SpecCodeCount && strcasecmp(SpecCodeNames[fcode], command) != 0;
		fcode++);
	  
	  if (fcode >= SpecCodeCount) fcode = -1;
	}	
  
	if (fcode < 0) 
	  printf ("Sorry, that was an invalid code\n");

  }	while ((--iter > 0) && (fcode < 0));  
  
  if (fcode < 0) {
    *errMsg = (char *) cancelledText ;
	return False;
  }	
  
  EnterText("Arguments (if necessary): ", parms, sizeof(parms));

  sprintf (command, "%s %s", SpecCodeNames[fcode], parms);
  return True;

}

Bool editExternalApp(char *command, int sze, char **errMsg)
{
  
  EnterText("Enter the shell command to call the application:\n", command, sze);
  return (*command != '\0');
}


Bool getEventTarget(char *data, int sze, short *target_sel, short canSelectOrder, 
                   int *xptr, int *yptr, int *button, int *state, char **errMsg)
{
  Window cur, tmp;
  XClassHint xch;
  XEvent xev;
  Bool tsel_defined;
  XTextProperty xtp;
  char **list;
  int i;
  char keyPressed;

  if (canSelectOrder) {
  	i = getMouseOrKeyEvent(&xev, btnTimeout, TargetSelCodes, &keyPressed, errMsg);
	
	if ( i == 2) {
  		// Key has been pressed
		*target_sel = 3;	
		*data = keyPressed;
		*(data+1) = '\0';
		return True;
	}			
	
  }	
  else
  	i = getMouseEvent(&xev, btnTimeout, errMsg);
             	

  
  
  if (i != 1) {
  	if (*errMsg == NULL) errMsg = (char **) timeoutText;
	return False;
  }

  //  Valid button press
  tmp = xev.xbutton.subwindow;
  if (tmp == None)
	cur = wnd;
  else {  
	cur = findClientWindow(tmp, -1);
	if (cur == None) cur = tmp;
  }  

  // res_name + ress_class myst exist and be unique
  tsel_defined = False;
  if (XGetClassHint(dpy, cur, &xch) && 
      countWindowsByWMClass(wnd, xch.res_name, xch.res_class) == 1) {
	tsel_defined = True;
	*target_sel = 0;	
	snprintf(data, sze, "%s:%s", xch.res_name, xch.res_class);
  } else
  if (XGetWMName(dpy, cur, &xtp) &&
      XTextPropertyToStringList(&xtp, &list, &i)) {
	  if (i > 0) {
		tsel_defined = True;
		*target_sel = 1;	
		strncpy(data, list[0], sze);
	  }	
  }  	 

  if (tsel_defined == False) {
	*target_sel = 2;	
	snprintf(data, sze, "0x%lx", (long)cur);		
  }	

  if (xptr != NULL && yptr != NULL)
  	  XTranslateCoordinates(dpy, xev.xbutton.root, cur,
	                               xev.xbutton.x_root,  xev.xbutton.y_root,
								   xptr, yptr, &tmp);
  if (button != NULL)
	*button = xev.xbutton.button;

  if (state != NULL)
	*state = xev.xbutton.state;
	
  return True;
}

Bool getEventType(int *type, char **errMsg)
{
  int sel;
  static const char *menuItems[] = {"Press", "Release", "Press and Release"};
	
  sel = DisplayMenu("Select the events to generate:",
                    menuItems, 3);
  if (sel < 0) {
    *errMsg = (char *) cancelledText ;
   return False;
 }
  *type = sel+1;  

  return True;
}

Bool getEventCount(int *countPtr, char **errMsg)
{
  char countStr[6], *next;
  int count;

  EnterText("Select event count (1 by default)\n", countStr, sizeof(countStr));
  if (*countStr == '\0') return True;

  count = strtol(countStr, &next, 0);
  if (*next=='\0' && count > 0) {
	*countPtr = count;
	return True;  
  }

  *errMsg = (char *) cancelledText ;
  return False;
}  

Bool editKeyEvent(char *command, int sze,
                  short *target_sel, short *type_modif, char **errMsg)
{
  XEvent xev;
 /*  char answ[4]; */
  int curSize, type;

/*
  EnterText("\nBring up the window to get key events,\ntype Y when ready: ", answ, sizeof(answ));
  if (toupper(answ[0]) != 'Y') {
	*errMsg = (char *) cancelledText;
	return False;
  }
*/  
  printf ("\nClick a window to receive key events,\nor press key: T-top wnd, B-bottom wnd, R-root wnd:\n");
  if (getEventTarget(command, sze, target_sel, 1,
                      NULL, NULL, NULL, NULL, errMsg) == False) {
	return False;
  }
	
  printf ("Press a key (key combination) to send\n");
  *type_modif = 1;			// Scan code
  if (getScanCodeEvent(&xev, keyTimeout, errMsg) != 1) {
	if (errMsg == NULL) *errMsg = (char *) timeoutText;
	return False;
  } 
  curSize = strlen(command);
  snprintf(command+curSize, sze-curSize, ";%d;%d", xev.xkey.keycode, xev.xkey.state);

  if (getEventType(&type, errMsg) == False) return False;
  curSize = strlen(command);
  snprintf(command+curSize, sze-curSize, ";%d", type);
  
  if (getEventCount(&type, errMsg)) {
	curSize = strlen(command);
	snprintf(command+curSize, sze-curSize, ";%d", type);
  }	

  return True;
}

Bool editButtonEvent(char *command, int sze, short *target_sel, char **errMsg)
{
  int curSize, type;
  int x, y;
  int state, button;
  char answ[4];

  EnterText("\nBring up the window to get the mouse click,\ntype Y when ready: ", answ, sizeof(answ));
  if (toupper(answ[0]) != 'Y') {
	*errMsg = (char *)cancelledText ;
	return False;
  }
  
  printf ("\nPerform a mouse click to be simulated\n");
  if (getEventTarget(command, sze, target_sel, 0,
           &x, &y, &button, &state, errMsg) == False) {
	*errMsg = (char *) cancelledText ;					  
	return False;
  }	
	
  curSize = strlen(command);
  snprintf(command+curSize, sze-curSize, ";%d;%d;%d;%d", x, y, button, state);

  if (getEventType(&type, errMsg) == False) return False;
  curSize = strlen(command);
  snprintf(command+curSize, sze-curSize, ";%d", type);
  
  if (getEventCount(&type, errMsg)) {
	curSize = strlen(command);
	snprintf(command+curSize, sze-curSize, ";%d", type);
  }	
  return True;
}

int GetFreePosition()
{
  int i;
  for (i=0; i<CommandCount && Commands[i].lineno <= (i+1); i++);
  if (i>=maxCodes) maxCodes = i+1;

  return i;
}


Bool editCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg)
{
  int pos, iter;
  int cmd_type;
  CmdDescr cmdNew;
  char command[257];
  Bool success;
  
  pos  = findCommandByCode(scanCode, modif);

  if (pos >= 0 && displayCommandEntry(pos, errMsg) &&
	  YesNo ("Overwrite", False) == False) {
	*errMsg = (char *) cancelledText;  return False;
  }	

  success = False;
  cmdNew.keycode = scanCode;
  cmdNew.modifier = modif;
  cmdNew.type_modifier = 0;
  cmdNew.target_selection = 0;
  cmdNew.command = NULL;

  iter = 3;
  
  do
  {
	cmd_type =DisplayMenu("Select command type", CommandTypeNames, 
							sizeof(CommandTypeNames) / sizeof(char *));
	if (cmd_type < 0) break;
  

    cmdNew.type = cmd_type;

	switch (cmd_type) 
	{
	  case CT_INTERNAL:
		success = editSpecFunction(command, sizeof(command), errMsg);
		break;

	  case  CT_EXTAPP:
		success = editExternalApp(command, sizeof(command), errMsg);
	    break;

	  case  CT_KEYEVENT:
		success = editKeyEvent(command, sizeof(command),
		            &(cmdNew.target_selection), &(cmdNew.type_modifier), errMsg);
		break;

	  case  CT_BTNEVENT:
		success = editButtonEvent(command, sizeof(command),
		            &(cmdNew.target_selection), errMsg);
		break;
	}  

	if (success) {
	  cmdNew.command = strdup(command);
	  printf ("\n");
	  success  = displayCommandLine(&cmdNew, errMsg);
	  if (success)
		success = YesNo ("Accept", True);
	}

  } while ((--iter > 0) && (success == False));

  if (success == False) goto TheEnd;

  if (pos >= 0) {
    cmdNew.lineno = Commands[pos].lineno;
    if (Commands[pos].command) free((char *)(Commands[pos].command));
  } else {
    pos = GetFreePosition();
    cmdNew.lineno = pos+1;
	CommandCount++;

    Commands = realloc(Commands, CommandCount*sizeof(CmdDescr));
    success = (Commands != NULL);
	if (success == False) goto TheEnd;
		  
	for (iter = CommandCount-1; iter > pos; iter--)
	  memcpy(&Commands[iter], &Commands[iter-1], sizeof(CmdDescr));	
  }	

  memcpy(&Commands[pos], &cmdNew, sizeof(CmdDescr));
  success = UpdateResource(resFileName, pos, errMsg);

TheEnd:
  if (success == False && cmdNew.command)
	free((char *)(cmdNew.command));
  return success;
}


Bool editCommand(char **errMsg)
{
  XEvent xev;
  short res;
  short cnt;
  
  cnt = count;

  do
  {
	printf ("----------- ADD/MODIFY hot key -------------\n");
    printf("Type a key combination you wish to add/modify\n");
	printf ("Quits ");
	if (cnt > 1)  printf("after %d entries, or ", cnt);
	printf ("if idle within %d secs\n", keyTimeout);
	res = getScanCodeEvent(&xev, keyTimeout, errMsg);
	if (res != 1)  break;
	editCommandByCode(xev.xkey.keycode, xev.xkey.state, errMsg);
	printf ("\n");
  }	while (cnt == 0 || --cnt > 0);

  if (res == 2)
	printf(timeoutBye);
  return (res != 0);	 
}


/* ---------------------------------------------------------------------- */
Bool deleteCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg)
{
  int pos;
  
  pos  = findCommandByCode(scanCode, modif);

  if (pos < 0) {
	*errMsg = "Key/key combination is not assigned";
	return False;
  }

  if (pos >= 0 && displayCommandEntry(pos, errMsg) &&
	  YesNo ("Delete", False) == False) {
	*errMsg = (char *) cancelledText;  return False;
  }	

  Commands[pos].type = CT_DELETED;

  if (UpdateResource(resFileName, pos, errMsg)) {
	int iter;

    CommandCount--;
	for (iter = pos; iter < CommandCount; iter++)
	  memcpy(&Commands[iter], &Commands[iter+1], sizeof(CmdDescr));	

	return True;
  }

  return False;
}


Bool deleteCommand(char **errMsg)
{
  XEvent xev;
  short res;
  short cnt;
  
  cnt = count;

  do
  {
	printf ("---------  DELETE  hot  key   ----------\n");
    printf("Type a key combination you wish to delete\n");
	printf ("Quits ");
	if (cnt > 1)  printf("after %d entries, or ", cnt);
	printf ("if idle within %d secs\n", keyTimeout);
	res = getScanCodeEvent(&xev, keyTimeout, errMsg);
	if (res != 1)  break;
	deleteCommandByCode(xev.xkey.keycode, xev.xkey.state, errMsg);
	printf ("\n");
  }	while (cnt == 0 || --cnt > 0);

  if (res == 2)
	printf(timeoutBye);
  return (res != 0);	 
}
