/*===================================================*/
/*         Hot key Processing                        */
/*---------------------------------------------------*/
/*       20.07.02                          M.G.      */
/*===================================================*/


#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xhkeys_conf.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>

#define _GNU_SOURCE
#include <getopt.h>

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "xhkeys.h"

/* system */
extern char **environ;

/* resources.c */
extern CmdDescr *Commands;
extern int CommandCount;
extern int maxCodes;
extern const char *LogModeCodes;
extern int logMode;		// 0 - none, 1 - error, 2 - all

Display *dpy = NULL;
Window  wnd = None;
Window  root_wnd = None;
Atom WMStateAtom, XHKeysPidAtom;
Bool debug = False;
Bool help = False;

static const char *resFileName = NULL;
static void (*oldTerminateHandler)(int) = NULL;
static void (*oldChildExitHandler)(int) = NULL;

static void terminateHandler(int sigCode);
static void childExitHandler(int sigCode);

void log_message(Bool error, char *msg)
{
  if (logMode == 2 || logMode == -1 || (logMode == 1 && error)) {
#ifdef USE_SYSLOG
	if (debug == False)
  	  syslog (error ? LOG_ERR : LOG_INFO, "%s\n", msg);

	if (debug || error)
#endif  	
	  fprintf (error ? stderr : stdout, "%s\n", msg);
  }	  
}

static Bool ProcessExtApp(const char * command)
{
  int pid;
  
  
  pid = fork();
  if (pid == 0) {
	char *shell, *sh_name;
	char * argv[4];
	char cmdcopy[512];
	
	shell = getenv("SHELL");
	if (shell == NULL)  shell = "/bin/sh";
  
	sh_name = strrchr(shell, '/');
	if (sh_name == NULL) sh_name = shell; else sh_name++;

	signal(SIGCHLD, childExitHandler);

	strncpy(cmdcopy, command, sizeof(cmdcopy));
	argv[0] = sh_name;
	argv[1] = "-c";
	argv[2] = cmdcopy;
	argv[3] = NULL;
	freopen("/dev/null", "w", stdout);
	freopen("/dev/null", "r", stdin);
	execve(shell, argv, environ);

	XBell(dpy, 50);
  }

  return True;
}


static void showHelp(void) {
  static const char *helpText[] = {
	"",
	"xhkeys ver. " VERSION_TEXT ", compiled " __DATE__ " " __TIME__,
	"Allows using keyboard extra keys with X server.",
	"(C) 2002, Michael Glickman <wmalms@yahoo.com>",
	"Web: www.geocities.com/wmalms",
	"",
	"Command line arguments:",
	"\t-d --debug        no fork, log to stdout", 
	"\t-f --file <name>  configuration file name", 
	"\t                  (default ~/.XHkeys)",
	"\t-h --help         this page", 
	"\t-l --log <.>      log: N-none, E-error, A-all", 
	"", 
	"Use xhkconf for configuration",
	"See xhkeys.html for details", 
	"", NULL };
	
  const char *helpLine;
  int i;
  
  for (i=0; (helpLine=helpText[i])!= NULL; i++)
	puts(helpLine);
	
	
	

}


static Bool ProcessCommandLine(int argc, char * const argv[])
{

  int result;
  
  static struct option long_opts[] =
  { { "file", required_argument, NULL, 'f'},
    { "debug", no_argument, NULL, 'd'},
    { "help", no_argument, NULL, 'h'},
    { "log", required_argument, NULL, 'l'},
    {NULL, 0, NULL,  0} 
  };

  
  resFileName = NULL;
  debug = False;
  help = False;

  logMode = -1;
  
  while ((result = getopt_long(argc, argv, "f:dhl:", long_opts, NULL)) != -1) {
	switch(result) {
	  case 'f':
		resFileName = optarg;
		break;	

	  case 'd':
		debug = True;
		break;	

	  case 'h':
		help = True;
		return True;

	  case 'l':
	  {
		char *tmp = strchr(LogModeCodes, toupper(*optarg));
		if (tmp == NULL) {
		  fprintf (stderr, "-- invalid log action\n");
		  return False;
		}  
		else
		logMode = (tmp-LogModeCodes);  
	  }
		break;	
		
	  case '?':
		return False;	
	}
  
  }
  
  return True;

}

static void terminate(void)
{
  int i;
  
  if (oldTerminateHandler != NULL)
	signal(SIGTERM, oldTerminateHandler);

  if (oldChildExitHandler != NULL)
	signal(SIGCHLD, oldChildExitHandler);

  for (i=0; i<CommandCount; i++) 
    XUngrabKey(dpy, Commands[i].keycode, Commands[i].modifier, wnd);


  DeinitSpecFunctions();
  FreeResources();

  if (XHKeysPidAtom != None) 
	XDeleteProperty(dpy, root_wnd, XHKeysPidAtom);

/*  if (wnd) XDestroyWindow(dpy, wnd); */

  if (dpy != NULL) XCloseDisplay(dpy);
}

static void terminateHandler(int sigCode)
{
/*
  terminate(False, 
#ifdef USE_SYSLOG
	"xkeys killed"
#else
	NULL
#endif  
  );
*/
  char msg[61];
  sprintf (msg, "xhkeys [pid %d] killed",  getpid());
  log_message(False, msg);

  terminate();
  exit(4);
}

// This will help to get rid of zombies
static void childExitHandler(int sigCode)
{
	wait(NULL);
}

int main(int argc, char * const argv[])
{
  XEvent xev;
  char *errMsg = NULL;
  const char * command;
  int i, keycode, type;
  unsigned int modif;
  Bool success = False; 
  pid_t pid;


  resFileName = NULL;
  
  if (!ProcessCommandLine(argc, argv) || help) {
  	showHelp();
	return help ? 0 : 16;
  }	  
 
  if (!debug) // && getpid() == tcgetpgrp(0))
  {
    if ((i=fork()) != 0) {
	  if (i == -1) errMsg = "Forking error!";
	         else success = True;
	  goto OutOfHere; 
  	} 
  }

  dpy = XOpenDisplay(NULL);
  if (dpy == NULL) {
	errMsg = "Can't open display";
	goto OutOfHere;
  }  

  root_wnd = DefaultRootWindow(dpy);
  if (root_wnd == None) {
	errMsg = "No root window";
	goto OutOfHere;
  }  

  wnd = root_wnd;
/*
  wnd = XCreateSimpleWindow(dpy, root_wnd, 0, 0, 10, 10, 1, 0, 0);
  if (wnd == None) {
	errMsg = "Couldn't create a window";
	goto OutOfHere;
  }  

  XMapWindow(dpy, wnd);
  XMoveWindow(dpy, wnd, -1000, -1000);
*/  

  if (!ProcessResources(resFileName, True, True, &errMsg)) goto OutOfHere;

  if (CommandCount == 0) {
  	errMsg = "No scan codes specified";
	goto OutOfHere;
  }

  if (CommandCount < maxCodes) 
    Commands = realloc(Commands, CommandCount*sizeof(CmdDescr));

  success = (Commands != NULL);  	
  if (!InitSpecFunctions(&errMsg)) goto OutOfHere;

  if (logMode < 0) logMode = 1;
  pid = getpid();

//#ifdef USE_SYSLOG
  {
	char msg[81];
	struct passwd * pwd = getpwuid(getuid());
	
	if (pwd == NULL) 
	  sprintf (msg, "xhkeys [pid %d] started", pid);
	else  
	  sprintf (msg, "xhkeys [pid %d] started by '%s'", pid, pwd->pw_name); 
	  
	log_message(False, msg);

  }	
//#endif

  XHKeysPidAtom = XInternAtom(dpy, "XHKeysPid", False);

  if (XHKeysPidAtom != None) {
	Atom actual_type;
	int actual_format;
	unsigned long nitems;
    unsigned long bytes_after;
    unsigned char *prop;

	// Kill old process if any
	XGetWindowProperty(dpy, root_wnd, XHKeysPidAtom, 0, sizeof(pid), False,
       AnyPropertyType, &actual_type, &actual_format, &nitems, &bytes_after, &prop);
	if (actual_type == XHKeysPidAtom) {
	  pid_t oldPid = *(pid_t *)prop; 
	  kill(oldPid, SIGTERM);
	  while (kill(oldPid, 0) == 0) sleep (0);
	} 
	  
	// Store current pid
    XChangeProperty(dpy, root_wnd, XHKeysPidAtom, XHKeysPidAtom, 8, PropModeReplace,
	                     (unsigned char *) &pid, sizeof(pid));

  }

  oldTerminateHandler = signal(SIGTERM, terminateHandler);
  oldChildExitHandler = signal(SIGCHLD, childExitHandler);
  
  WMStateAtom = XInternAtom(dpy, "WM_STATE", True);
  
  for (i =0; i < CommandCount; i++) {
    XGrabKey(dpy, Commands[i].keycode, Commands[i].modifier, wnd, False, GrabModeAsync, GrabModeAsync);
  }	

  // From this point we cant fail!
  // XSelectInput(dpy, wnd, KeyPressMask);

  // Now we can start using 'success' for a slightly different purpose
  for(;;) {
	XNextEvent(dpy, &xev);
	if (xev.type != KeyPress) continue;

	success = False;
	keycode = xev.xkey.keycode;
	modif = xev.xkey.state;

	for (i =0; i < CommandCount &&
	      (Commands[i].keycode != keycode || Commands[i].modifier != modif); i++) ;
	if (i>=CommandCount) continue;

	  
	command = Commands[i].command;

	switch(type = Commands[i].type)
	{ 
	  case CT_INTERNAL:
		i = ProcessSpecFunction(command);
		if (i<0) goto OutOfLoop;
		success = (i>0);
		break;
		
	  case CT_EXTAPP:
		success = ProcessExtApp(command);
		break;	
		
	  case CT_KEYEVENT:
		success = ProcessWndKey(command, Commands[i].target_selection,
		                                 Commands[i].type_modifier);
		break;

	  case CT_BTNEVENT:
		success = ProcessWndButton(command, Commands[i].target_selection);
		break;
	}		

    if (success == False) XBell(dpy, 50);

  } 

  
OutOfLoop:
  terminate();

//#ifdef USE_SYSLOG
  {
	char msg[61];
	sprintf (msg, "xhkeys [pid %d] finished", pid);
	log_message(False, msg);
  }	
//#endif  
  success = True;

OutOfHere:
  if (errMsg) log_message(!success, errMsg);
  return success;
}
