
typedef struct {
  unsigned int lineno;
  unsigned int keycode;
  unsigned int modifier;
  int type;
  short type_modifier;	      // Key: 0 - keysym, 1 - scan code
  short target_selection;	  // Select target window by:
                              //    0 - WM_CLASS, 1 - by Title, 2 - ID
  const char *command;    // Command or window
} CmdDescr; 

enum CommandType {
  CT_INTERNAL,
  CT_EXTAPP,
  CT_KEYEVENT,
  CT_BTNEVENT,
  CT_DELETED=99
};

#define CommandTypeCount 4

enum SpecCodes {

#ifdef CDROM_SUPPORT
		SPEC_CDPLAYPAUSE, SPEC_CDPLAY, 
		SPEC_CDPAUSE,     SPEC_CDSTOP,
        SPEC_CDLOADEJECT, SPEC_CDLOAD, 
		SPEC_CDEJECT,     SPEC_CDSLOT,
#endif
#ifdef MIXER_SUPPORT
		SPEC_SNDVOLUME,   SPEC_SNDBALANCE,
#endif
		SPEC_WNDROTATE,   SPEC_QUIT};



// Spec func
Bool InitSpecFunctions(char **errMsg);
void DeinitSpecFunctions(void);
int ProcessSpecFunction(const char *command);

// Resources
Bool ProcessResources(const char *fileName, Bool getGlobal, Bool getCmdLines, char **errMsg);
Bool UpdateResource(const char *fileName, int pos, char **errMsg);
void FreeResources(void);


// Key and button events
Bool ProcessWndKey(const char * command, short target_sel, short type_modif);
Bool ProcessWndButton(const char * command, short target_sel);


// xhkconf.c
void xhkc_terminate(void);

// xhkccodes.c
Bool showScanCodes(char **errMsg);
Bool showMouseEventInfos(char **errMsg);
short getScanCodeEvent(XEvent *xevPtr, unsigned int timeout, char **errMsg);
short getMouseEvent(XEvent *xevPtr, unsigned int timeout, char **errMsg);
short getMouseOrKeyEvent(XEvent *xevPtr, unsigned int timeout,
                        const char *keysAllowed, char *keyPressed,  char **errMsg);

// xhkccmds.c
Bool displayCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg);
Bool displayAllCommands(char **errMsg);
Bool displayCommand(char **errMsg);

Bool editCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg);
Bool editCommand(char **errMsg);
Bool deleteCommandByCode(unsigned int scanCode, unsigned int modif, char **errMsg);
Bool deleteCommand(char **errMsg);

// common.c

typedef Bool (*LookupWindowHandler)(Window wnd, const void *par);
Window lookupWindow(Window cur, int dir, LookupWindowHandler handler, const void *par);
Window findWindowByWMClass(Window main_wnd, const char *app_name, const char *class_name);
Window findWindowByWMState(Window main_wnd, int dir, long state);
Window findVisibleWindow(Window main_wnd, int dir);
Window findClientWindow(Window main_wnd, int dir);
Window findWindowByTitle(Window main_wnd,  const char *title);
Window findWindowById(Window main_wnd, Window id);
int   countWindowsByWMClass(Window main_wnd, const char *app_name, const char *class_name);
Bool  GetWindowState(Window window, long *state);

#ifndef HAVE_STRSEP
char *strsep(char **stringp, const char *delim);
#endif
